//**********************************************************************
// KOLMonthCalendar - -  MonthCalendar Win32 Common Controls
//
// 2006 ECM
// version 0.02   18.02.06 (beta)
//**********************************************************************
unit KOLMonthCalendar;
{* KOLMontCalendar - A month calendar control implements a calendar-like 
user interface. This provides the user with a very intuitive and recognizable 
method of entering or selecting a date. The control also provides the 
application with the means to obtain and set the date information in the
control using existing data types.}
interface
uses Windows, KOL, Messages;

//{$DEFINE AUTO_ADJUST_SIZE}

{* ====== MONTHCAL CONTROL ========= }

const
  MONTHCAL_CLASS          = 'SysMonthCal32';

const
  // Message constants
  MCM_FIRST             = $1000;
  MCM_GETCURSEL         = MCM_FIRST + 1;
  MCM_SETCURSEL         = MCM_FIRST + 2;
  MCM_GETMAXSELCOUNT    = MCM_FIRST + 3; //+
  MCM_SETMAXSELCOUNT    = MCM_FIRST + 4; //+
  MCM_GETSELRANGE       = MCM_FIRST + 5;
  MCM_SETSELRANGE       = MCM_FIRST + 6;
  MCM_GETMONTHRANGE     = MCM_FIRST + 7;
  MCM_SETDAYSTATE       = MCM_FIRST + 8;
  MCM_GETMINREQRECT     = MCM_FIRST + 9;
  MCM_SETCOLOR          = MCM_FIRST + 10; //+
  MCM_GETCOLOR          = MCM_FIRST + 11; //+
  MCM_SETTODAY          = MCM_FIRST + 12;
  MCM_GETTODAY          = MCM_FIRST + 13;
  MCM_HITTEST           = MCM_FIRST + 14;
  MCM_SETFIRSTDAYOFWEEK = MCM_FIRST + 15; //+
  MCM_GETFIRSTDAYOFWEEK = MCM_FIRST + 16; //+
  MCM_GETRANGE          = MCM_FIRST + 17; // +
  MCM_SETRANGE          = MCM_FIRST + 18; // +
  MCM_GETMONTHDELTA     = MCM_FIRST + 19; //+
  MCM_SETMONTHDELTA     = MCM_FIRST + 20; //+
  MCM_GETMAXTODAYWIDTH  = MCM_FIRST + 21;
  MCM_SETUNICODEFORMAT  = CCM_SETUNICODEFORMAT;
  MCM_GETUNICODEFORMAT  = CCM_GETUNICODEFORMAT;

  // Hit test flags
  MCHT_TITLE            = $00010000;
  MCHT_CALENDAR         = $00020000;
  MCHT_TODAYLINK        = $00030000;
  MCHT_NEXT             = $01000000;  // these indicate that hitting
  MCHT_PREV             = $02000000;  // here will go to the next/prev month
  MCHT_NOWHERE          = $00000000;
  MCHT_TITLEBK          = MCHT_TITLE;
  MCHT_TITLEMONTH       = MCHT_TITLE or $0001;
  MCHT_TITLEYEAR        = MCHT_TITLE or $0002;
  MCHT_TITLEBTNNEXT     = MCHT_TITLE or MCHT_NEXT or $0003;
  MCHT_TITLEBTNPREV     = MCHT_TITLE or MCHT_PREV or $0003;
  MCHT_CALENDARBK       = MCHT_CALENDAR;
  MCHT_CALENDARDATE     = MCHT_CALENDAR or $0001;
  MCHT_CALENDARDATENEXT = MCHT_CALENDARDATE or MCHT_NEXT;
  MCHT_CALENDARDATEPREV = MCHT_CALENDARDATE or MCHT_PREV;
  MCHT_CALENDARDAY      = MCHT_CALENDAR or $0002;
  MCHT_CALENDARWEEKNUM  = MCHT_CALENDAR or $0003;

  // Color codes
  MCSC_BACKGROUND       = 0;   // the background color (between months)
  MCSC_TEXT             = 1;   // the dates
  MCSC_TITLEBK          = 2;   // background of the title
  MCSC_TITLETEXT        = 3;
  MCSC_MONTHBK          = 4;   // background within the month cal
  MCSC_TRAILINGTEXT     = 5;   // the text color of header & trailing days

  // Notification codes
  MCN_FIRST             = 0-750;
  MCN_SELCHANGE         = MCN_FIRST + 1;
  MCN_GETDAYSTATE       = MCN_FIRST + 3;
  MCN_SELECT            = MCN_FIRST + 4;

  // Style flags
  MCS_DAYSTATE          = $0001;
  MCS_MULTISELECT       = $0002; // +
  MCS_WEEKNUMBERS       = $0004; // +
  MCS_NOTODAY_PRE_IE4   = $0008; // +
  MCS_NOTODAYCIRCLE     = $0008; // +
  MCS_NOTODAY           = $0010; // +

  GMR_VISIBLE           = 0;       // visible portion of display
  GMR_DAYSTATE          = 1;       // above plus the grayed out parts of
                                   // partially displayed months

type
  // bit-packed array of "bold" info for a month
  // if a bit is on, that day is drawn bold
  MONTHDAYSTATE = DWORD;
  PMonthDayState = ^TMonthDayState;
  TMonthDayState = MONTHDAYSTATE;

  MCHITTESTINFO = packed record
    cbSize: UINT;
    pt: TPoint;
    uHit: UINT;      // out param
    st: TSystemTime;
  end;

  PMCHitTestInfo = ^TMCHitTestInfo;
  TMCHitTestInfo = MCHITTESTINFO;

  // MCN_SELCHANGE is sent whenever the currently displayed date changes
  // via month change, year change, keyboard navigation, prev/next button
  tagNMSELCHANGE = packed record
    nmhdr: TNmHdr;  // this must be first, so we don't break WM_NOTIFY
    stSelStart: TSystemTime;
    stSelEnd: TSystemTime;
  end;
  PNMSelChange = ^TNMSelChange;
  TNMSelChange = tagNMSELCHANGE;

  TCalDayOfWeek = (dowMonday, dowTuesday, dowWednesday, dowThursday,
    dowFriday, dowSaturday, dowSunday, dowLocaleDefault);

  TMonthCalendarOption = (mcoWeekNumbers, mcoTodayCircle, mcoToday, mcoMultiSelect);
  TMonthCalendarOptions = set of TMonthCalendarOption;

const
  CalendarFlags: Array [TMonthCalendarOption] of Integer = (
    MCS_WEEKNUMBERS, not MCS_NOTODAYCIRCLE, not MCS_NOTODAY, MCS_MULTISELECT
  );

type

  // MCN_GETDAYSTATE is sent for MCS_DAYSTATE controls whenever new daystate
  // information is needed (month or year scroll) to draw bolding information.
  // The app must fill in cDayState months worth of information starting from
  // stStart date. The app may fill in the array at prgDayState or change
  // prgDayState to point to a different array out of which the information
  // will be copied. (similar to tooltips)
  tagNMDAYSTATE = packed record
    nmhdr: TNmHdr;  // this must be first, so we don't break WM_NOTIFY
    stStart: TSystemTime;
    cDayState: Integer;
    prgDayState: PMonthDayState; // points to cDayState TMONTHDAYSTATEs
  end;
  PNMDayState = ^TNMDayState;
  TNMDayState = tagNMDAYSTATE;

  // MCN_SELECT is sent whenever a selection has occured (via mouse or keyboard)
  NMSELECT = tagNMSELCHANGE;
  PNMSelect = ^TNMSelect;
  TNMSelect = NMSELECT;

  TDateRange = packed record
    MinDate: TDateTime;
    MaxDate: TDateTime;
  end;

  TOnGetMonthInfo = procedure( Sender: PControl; Month: Cardinal;
    var MonthBoldInfo: Cardinal) of object;

  TOnSelect = procedure (Sender: PControl; const SelStart, SelEnd: TDateTime) of object;

  PMonthCalData = ^TMonthCalData;
  TMonthCalData = packed record
    fMinDate:        TDateTime;
    fMaxDate:        TDateTime;
    fSelStart:       Cardinal;
    fSelEnd:         Cardinal;
    fOnSelect:       TOnSelect;
    fOnGetMonthInfo: TOnGetMonthInfo;
  end;

  PMonthCalendar    = ^TMonthCalendar;
  TKOLMonthCalendar = PMonthCalendar;

  TMonthCalendar = object(TControl)
  {* TKOLMonthCalendar is stand-alone calendar in which a user can select a date or range of dates.}
  private
    function  GetOptVal (const Index: Integer): Boolean;
    procedure SetOptVal (const Index: Integer; const Value: Boolean);
    function  GetIntVal (const Index: Integer): Integer;
    procedure SetIntVal (const Index: Integer; const Value: Integer);
    function  GetColVal (const Index: Integer): TColor;
    procedure SetColVal (const Index: Integer; const Value: TColor);
    function  GetDayOfWeek: TCalDayOfWeek;
    procedure SetDayOfWeek (const Value: TCalDayOfWeek);
    function  GetDateVal(const Index: Integer): TDateTime;
    procedure SetDateVal(const Index: Integer; const Value: TDateTime);
    function  GetMR(const Index: Integer): TDateRange;
    procedure SetSelChange(const Value: TOnEvent);
    function  GetOnGetMInfo: TOnGetMonthInfo;
    procedure SetOnGetMInfo(const Value: TOnGetMonthInfo);
    function  GetOnSelect: TOnSelect;
    procedure SetOnSelect(const Value: TOnSelect);
    function  GetMaxDate: TDateTime;
    function  GetMinDate: TDateTime;
    procedure SetMaxDate(const Value: TDateTime);
    procedure SetMinDate(const Value: TDateTime);
    function  GetSelEnd: TDateTime;
    function  GetSelStart: TDateTime;
    procedure SetSelEnd(const Value: TDateTime);
    procedure SetSelStart(const Value: TDateTime);
  protected
  public
    procedure AdjustSize;
    {* Sets the minimum size required to display a full month in a
       month calendar control.}
    procedure BoldDays(Days: array of LongWord; var MonthBoldInfo: LongWord);
    {* Encodes a set of dates for use in the OnGetMonthInfo event handler.}
    procedure SetBoldDays(const MonthBoldInfo: Array of LongWord);
    {* Sets the day states for all months that are currently visible within
       a month calendar control..}
    function  GetSelRange: TDateRange;
    {* Gets the selection for a month calendar control to a given date range.
       Apply to a month calendar control that USES the MCS_MULTISELECT style}
    procedure SetSelRange(const ASelStart, ASelEnd: TDateTime);
    {* Sets the selection for a month calendar control to a given date range.
       Apply to a month calendar control that USES the MCS_MULTISELECT style}
    function  GetRange: TDateRange;
    {* Gets minimum and maximum allowable dates for a month calendar control. }
    procedure SetRange(const AMinDate, AMaxDate: TDateTime);
    {* Sets minimum and maximum allowable dates for a month calendar control. }
    function  HitTest(const X,Y: Integer; var HitDate: TDateTime): Integer;
    {* Determines which portion of a month calendar control is at a given point
       on the screen.
       Return: Sets values in MCHT_XXXXX constants:
|<pre>
  MCHT_TITLE            - The given point was over a month's title.
  MCHT_CALENDAR         - The given point was within the calendar.
  MCHT_TODAYLINK        - The given point was on the "today" link at the bottom of the month
                          calendar control.
  MCHT_NEXT             - The given point is in an area that will cause the month calendar to scroll
                          its display to the next month or set of months. This flag is used to modify
                          other hit test flags.
  MCHT_PREV             - The given point is in an area that will cause the month calendar to scroll
                          its display to the previous month or set of months. This flag is used to
                          modify other hit test flags.
  MCHT_NOWHERE          - The given point was not on the month calendar control, or it was in an
                          inactive portion of the control.
  MCHT_TITLEBK          - The given point was over the background of a month's title.
  MCHT_TITLEMONTH       - The given point was in a month's title bar, over a month name.
  MCHT_TITLEYEAR        - The given point was in a month's title bar, over the year value.
  MCHT_TITLEBTNNEXT     - The given point was over the button at the top right corner of the control.
                          If the user clicks here, the month calendar will scroll its display to the
                          next month or set of months.
  MCHT_TITLEBTNPREV     - The given point was over the button at the top left corner of the control.
                          If the user clicks here, the month calendar will scroll its display to the
                          previous month or set of months.
  MCHT_CALENDARBK       - The given point was in the calendar's background.
  MCHT_CALENDARDATE     - The given point was on a particular date within the calendar. The HitDate variable
                          of TMonthCalendar.HitTest function is set to the date at the given point.
  MCHT_CALENDARDATENEXT - The given point was over a date from the next month (partially displayed at
                          the end of the currently displayed month). If the user clicks here, the month
                          calendar will scroll its display to the next month or set of months.
  MCHT_CALENDARDATEPREV - The given point was over a date from the previous month (partially displayed at
                          the end of the currently displayed month). If the user clicks here, the month
                          calendar will scroll its display to the previous month or set of months.
  MCHT_CALENDARDAY      - The given point was over a day abbreviation ("Fri", for example). The HitDate
                          variable of TMonthCalendar.HitTest function is set to the corresponding date
                          in the top row.
  MCHT_CALENDARWEEKNUM  - The given point was over a week number (WeekNumbers=TRUE style only).
                          The HitDate variable of TMonthCalendar.HitTest function is set to the corresponding
                          date in the leftmost column.
|</pre>
    }
    property WeekNumbers:     Boolean index 0 read GetOptVal write SetOptVal;
    {* The month calendar control will display week numbers (1-52) to the left
       of each row of days. Week 1 is defined as the first week that contains
       at least four days.}
    property ShowTodayCircle: Boolean index 1 read GetOptVal write SetOptVal;
    {* The month calendar control will circle the "today" date}
    property ShowToday:       Boolean index 2 read GetOptVal write SetOptVal;
    {* The month calendar control will display the "today" date at the
       bottom of the control}
    {= }
    property MultiSelect:     Boolean index 3 read GetOptVal;
    {* The month calendar will allow the user to select a range of dates
       within the control. By default, the maximum range is one week.}
    {= }
    property FirstDayOfWeek:  TCalDayOfWeek read GetDayOfWeek write SetDayOfWeek;
    {* First day of the week for a month calendar control. }
    property BackColor:       TColor index 0 read GetColVal write SetColVal;
    {* Background color displayed between months }
    property MonthBackColor:  TColor index 1 read GetColVal write SetColVal;
    {* Background color displayed within the month}
    property TextColor:       TColor index 2 read GetColVal write SetColVal;
    {* Color used to display text within a month}
    property TitleBackColor:  TColor index 3 read GetColVal write SetColVal;
    {* Background color displayed in the calendar's title}
    property TitleTextColor:  TColor index 4 read GetColVal write SetColVal;
    {* Color used to display text within the calendar's title}
    property TrailingTextColor:  TColor index 5 read GetColVal write SetColVal;
    {* Color used to display header day and trailing day text}
    property MaxSelectRange: Integer index 1 read GetIntVal write SetIntVal;
    {* Maximum number of days that can be selected in a month calendar control.
       Set this field will fail if applied to a month calendar control that
       does not use the MCS_MULTISELECT style}
    property UniCodeFormat: Integer index 2 read GetIntVal write SetIntVal;
    {* Unicode character format flag for the control. If this value is nonzero,
       the control is using Unicode characters. If this value is zero, the
       control is using ANSI characters}
    property MonthDelta: Integer index 3 read GetIntVal write SetIntVal;
    {* Scroll rate for a month calendar control. The scroll rate is the number
       of months that the control moves its display when the user clicks a
       scroll button.}
    //property SelRange:    TDateRange read GetSelRange write SetSelRange;
    {* Sets the selection for a month calendar control to a given date range.
       Apply to a month calendar control that USES the MCS_MULTISELECT style}
    property MinDate:     TDateTime read GetMinDate write SetMinDate;
    {* Minimum allowable dates for a month calendar control. Equivalent of
       MinMaxRange.MinDate }
    property MaxDate:     TDateTime read GetMaxDate write SetMaxDate;
    {* Maximum allowable dates for a month calendar control. Equivalent of
       MinMaxRange.MaxDate }
    property Date:        TDateTime  index 0 read GetDateVal write SetDateVal;
    {* Sets the currently selected date for a month calendar control.
       If the specified date is not in view, the control updates the display
       to bring it into view. Apply to a month calendar control that NOT uses
       the MCS_MULTISELECT style}
    property SelStart:    TDateTime read GetSelStart write SetSelStart;
    property SelEnd:      TDateTime read GetSelEnd write SetSelEnd;

    property Today:       TDateTime  index 1 read GetDateVal write SetDateVal;
    {* "Today" selection for a month calendar control. }
    property MonthRange:  TDateRange index 0 read GetMR;
    {* Retrieves date information that represents the high and low limits of
       a month calendar control's display. Include preceding and trailing months
       of visible range that are only partially displayed. }
    property MonthRangeVisible:  TDateRange index 1 read GetMR;
    {* Retrieves date information that represents the high and low limits of
       a month calendar control's display. Include only those months that are
       entirely displayed. }
    property OnSelChange: TOnEvent read fOnSelChange write SetSelChange;
    {* Sent by a month calendar control when the currently selected date or
       range of dates changes. Simple notify}
    property OnGetMonthInfo: TOnGetMonthInfo read GetOnGetMInfo write SetOnGetMInfo;
    {* Called whenever a new month is displayed in the calendar.Use the
       OnGetMonthInfo event to initialize the display properties of a month.
       In particular, OnGetMonthInfo allows applications to bold individual
       days in the calendar (such as holidays).}
    property OnSelect: TOnSelect read GetOnSelect write SetOnSelect;
    {* Sent by a month calendar control when the currently selected date or
       range of dates changes. Include selection range}
  end;

function NewMonthCalendar(AParent : PControl; Options : TMonthCalendarOptions) : PMonthCalendar;
function WndProcCalSelChange(Sender : PControl; var Msg:TMsg; var Rslt:Integer):Boolean;
function WndProcCalDayState(Sender : PControl; var Msg:TMsg; var Rslt:Integer):Boolean;

function DbgLog(const Str: String): Boolean;

implementation
type

  NMHDR = packed record
    hwndFrom: HWND;
    idFrom: UINT;
    code: Integer;
  end;
  TNMHdr = NMHDR;
  PNMHdr = ^TNMHdr;

  NMDATETIMECHANGE = packed record
    nmhdr: TNmHdr;
    dwFlags: DWORD;         // GDT_VALID or GDT_NONE
    st: TSystemTime;        // valid iff dwFlags = GDT_VALID
  end;

  PNMDateTimeChange = ^TNMDateTimeChange;
  TNMDateTimeChange = NMDATETIMECHANGE;

function DbgLog(const Str: String): Boolean;
begin
  LogFileOutput('.\deb_mc.log',Str);
  Result := TRUE;
end;


function NewMonthCalendar(AParent : PControl; Options: TMonthCalendarOptions):PMonthCalendar;
var
  Flags: Integer;
begin
  Flags := MakeFlags(@Options, CalendarFlags);// or MCS_DAYSTATE;// -    OnGetMonthInfo
  Assert(DbgLog(Format('NewMonthCalendar(%.8x,%.8x)',[AParent,Flags])));
  Result := PMonthCalendar(_NewCommonControl( AParent,MONTHCAL_CLASS, WS_CHILD or WS_VISIBLE or Flags, FALSE, nil));
  //   . .    TControl.Destroy
  GetMem(Result.fCustomData,SizeOf(TMonthCalData));
  FillChar(Result.fCustomData^,SizeOf(TMonthCalData),0);
end;

function WndProcCalSelChange(Sender : PControl; var Msg:TMsg; var Rslt:Integer):Boolean;
var
  NMSC:PNMSelChange;
  SelRange: TDateRange;
  MCD: PMonthCalData;
  SS,SE: Cardinal;
begin
  Result := False;
  if Msg.message=WM_NOTIFY then
  begin
    NMSC := PNMSelChange(Msg.lParam);
    MCD := PMonthCalData(Sender.CustomData);
    with NMSC^,MCD^ do begin
      if (nmhdr.code = MCN_SELECT) or (nmhdr.code = MCN_SELCHANGE) then begin
        SystemTime2DateTime(stSelStart, SelRange.MinDate);
        SystemTime2DateTime(stSelEnd, SelRange.MaxDate);
        SS := Trunc(SelRange.MinDate); SE := Trunc(SelRange.MaxDate);
        if (SS <> fSelStart) or (SE <> fSelEnd) then begin
          fSelStart := SS; fSelEnd := SE;
          if Assigned(fOnSelect) then fOnSelect(Sender,SelRange.MinDate,SelRange.MaxDate);
          if Assigned(Sender.OnSelChange) then Sender.OnSelChange(Sender);
        end;
        Result := TRUE;
      end;
    end;
  end;
end;

function WndProcCalDayState(Sender : PControl; var Msg:TMsg; var Rslt:Integer):Boolean;
var
  NMDS:PNMDayState;
  pBoldInfo: PMonthDayState;
  i: Integer;
  Month: Cardinal;
begin
  Result := False;
  if Msg.message=WM_NOTIFY then
  begin
    NMDS:=PNMDayState(Msg.lParam);
    with NMDS^ do begin
      if nmhdr.code = MCN_GETDAYSTATE then begin
        FillChar(prgDayState^, cDayState * SizeOf(TMonthDayState), 0);
        if Assigned(PMonthCalData(Sender.CustomData).fOnGetMonthInfo) then begin
          pBoldInfo := prgDayState;
          for i := 0 to cDayState - 1 do begin
            Month := stStart.wMonth + i;
            if Month > 12 then Month := Month - 12;
            PMonthCalData(Sender.CustomData).fOnGetMonthInfo(Sender,Month,pBoldInfo^);
            Inc(pBoldInfo);
          end;
        end;
        Result := TRUE;
      end;
    end;
  end;
end;

{ TMonthCalendar }

function TMonthCalendar.GetOptVal(const Index: Integer): Boolean;
begin
  UpdateWndStyles();
  if CalendarFlags[TMonthCalendarOption(Index)] > 0 then
    Result := (Style and CalendarFlags[TMonthCalendarOption(Index)]) <> 0
  else
    Result := (Style and not CalendarFlags[TMonthCalendarOption(Index)]) <> 0
end;

procedure TMonthCalendar.AdjustSize;
var
  R: TRect;
  tdWidth: Integer;
begin
  Perform(MCM_GETMINREQRECT,0,Integer(@R));
  if ShowToday then begin
    tdWidth := Perform(MCM_GETMAXTODAYWIDTH,0,0);
    if tdWidth > R.Right then R.Right := tdWidth;
  end;
  SetSize(R.Right,R.Bottom);
end;


procedure TMonthCalendar.BoldDays(Days: array of LongWord; var MonthBoldInfo: LongWord);
var
  i: Integer;
begin
  MonthBoldInfo := 0;
  for i := Low(Days) to High(Days) do
    if (Days[i] > 0) and (Days[i] < 32) then
      MonthBoldInfo := MonthBoldInfo or ($00000001 shl (Days[i] - 1));
end;

procedure TMonthCalendar.SetOptVal(const Index: Integer;
  const Value: Boolean);
var
  tmpVal: Boolean;
  Flag  : DWORD;
begin
  UpdateWndStyles();
  if CalendarFlags[TMonthCalendarOption(Index)] > 0 then begin
    tmpVal := Value;
    Flag   := CalendarFlags[TMonthCalendarOption(Index)];
  end else begin
    tmpVal := not Value;
    Flag   := not CalendarFlags[TMonthCalendarOption(Index)];
  end;
  if tmpVal then Style := Style or Flag
  else Style := Style and (not Flag);
  {$IFDEF AUTO_ADJUST_SIZE}
  AdjustSize;
  {$ENDIF}
end;

function TMonthCalendar.GetDayOfWeek: TCalDayOfWeek;
begin
  Result := TCalDayOfWeek(GetIntVal(0) and $f);
end;

procedure TMonthCalendar.SetDayOfWeek(const Value: TCalDayOfWeek);
var
  DOWFlag: Integer;
  A: array[0..1] of char;
begin
  if Value = dowLocaleDefault then begin
    GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_IFIRSTDAYOFWEEK, A, SizeOf(A));
    DOWFlag := Ord(A[0]) - Ord('0');
  end else DOWFlag := Ord(Value);
  Perform(MCM_SETFIRSTDAYOFWEEK,0,DOWFlag);
end;

function TMonthCalendar.GetIntVal(const Index: Integer): Integer;
const
  Msg_Val: Array[0..3] of DWORD = (
    MCM_GETFIRSTDAYOFWEEK,
    MCM_GETMAXSELCOUNT,
    MCM_GETUNICODEFORMAT,
    MCM_GETMONTHDELTA
  );
begin
  Result := Perform(Msg_Val[Index],0,0);
end;

procedure TMonthCalendar.SetIntVal(const Index, Value: Integer);
const
  Msg_Val: Array[0..3] of DWORD = (
    MCM_SETFIRSTDAYOFWEEK,
    MCM_SETMAXSELCOUNT,
    MCM_SETUNICODEFORMAT,
    MCM_SETMONTHDELTA
  );
begin
  Perform(Msg_Val[Index],Value,0);
end;

const
  CalColors: Array[0..5] of Integer = (
    MCSC_BACKGROUND, MCSC_MONTHBK, MCSC_TEXT, MCSC_TITLEBK,
    MCSC_TITLETEXT, MCSC_TRAILINGTEXT );

function TMonthCalendar.GetColVal(const Index: Integer): TColor;
begin
  Result := Perform(MCM_GETCOLOR,CalColors[Index],0);
end;

procedure TMonthCalendar.SetColVal(const Index: Integer;
  const Value: TColor);
begin
  Perform(MCM_SETCOLOR,CalColors[Index],Color2RGB(Value));
end;

function TMonthCalendar.GetRange: TDateRange;
var
  SysTimeArray: Array[0..1] of TSystemTime;
  Flag: DWORD;
begin
  FillChar(SysTimeArray[0],SizeOf(SysTimeArray),0);
  Flag := Perform(MCM_GETRANGE,0,Integer(@SysTimeArray[0]));
  if LongBool(Flag and GDTR_MIN) then
    SystemTime2DateTime(SysTimeArray[0],Result.MinDate);
  if LongBool(Flag and GDTR_MAX) then
    SystemTime2DateTime(SysTimeArray[1],Result.MaxDate);
  PMonthCalData(fCustomData).fMinDate := Result.MinDate;
  PMonthCalData(fCustomData).fMaxDate := Result.MaxDate;
end;

procedure TMonthCalendar.SetRange(const AMinDate, AMaxDate: TDateTime);
var
  SysTimeArray: Array[0..1] of TSystemTime;
  Flag: Integer;
begin
  FillChar(SysTimeArray[0],SizeOf(SysTimeArray),0);
  Flag := 0;
  if AMinDate <> 0 then begin
    DateTime2SystemTime(AMinDate,SysTimeArray[0]);
    Flag := GDTR_MIN;
  end;
  if AMaxDate <> 0 then begin
    DateTime2SystemTime(AMaxDate,SysTimeArray[1]);
    Flag := Flag or GDTR_MAX;
  end;
  Perform(MCM_SETRANGE,Flag,Integer(@SysTimeArray[0]));
  PMonthCalData(fCustomData).fMinDate := AMinDate;
  PMonthCalData(fCustomData).fMaxDate := AMaxDate;
end;

function TMonthCalendar.GetSelRange: TDateRange;
var
  SysTimeArray: Array[0..1] of TSystemTime;
begin
  Result.MinDate := 0;
  Result.MaxDate := 0;
  if LongBool(Perform(MCM_GETSELRANGE,0,Integer(@SysTimeArray[0]))) then begin
    SystemTime2DateTime(SysTimeArray[0],Result.MinDate);
    SystemTime2DateTime(SysTimeArray[1],Result.MaxDate);
  end;
  PMonthCalData(fCustomData).fSelStart := Trunc(Result.MinDate);
  PMonthCalData(fCustomData).fSelEnd := Trunc(Result.MaxDate);
end;

procedure TMonthCalendar.SetSelRange(const ASelStart, ASelEnd: TDateTime);
var
  SysTimeArray: Array[0..1] of TSystemTime;
begin
  FillChar(SysTimeArray[0],SizeOf(SysTimeArray),0);
  if ASelStart <> 0 then DateTime2SystemTime(ASelStart,SysTimeArray[0]);
  if ASelEnd <> 0 then DateTime2SystemTime(ASelEnd,SysTimeArray[1]);
  Perform(MCM_SETSELRANGE,0,Integer(@SysTimeArray[0]));
  PMonthCalData(fCustomData).fSelStart := Trunc(ASelStart);
  PMonthCalData(fCustomData).fSelEnd := Trunc(ASelEnd);
end;

function TMonthCalendar.GetDateVal(const Index: Integer): TDateTime;
const
  Msg_Val: Array[0..1] of DWORD = (
    MCM_GETCURSEL,
    MCM_GETTODAY
  );
var
  ST: TSystemTime;
begin
  Perform(Msg_Val[Index],0,Integer(@ST));
  SystemTime2DateTime(ST,Result);
end;

procedure TMonthCalendar.SetDateVal(const Index: Integer;
  const Value: TDateTime);
const
  Msg_Val: Array[0..1] of DWORD = (
    MCM_SETCURSEL,
    MCM_SETTODAY
  );
var
  ST: TSystemTime;
begin
  Assert(DbgLog('Date='+Int2Str(Trunc(Value))));
  DateTime2SystemTime(Value,ST);
  Assert(DbgLog('SD='+Int2Str(ST.wDay)+'/'+Int2Str(ST.wMonth)+'/'+Int2Str(ST.wYear)));
  Perform(Msg_Val[Index],0,Integer(@ST));
end;

function TMonthCalendar.GetMR(const Index: Integer): TDateRange;
var
  SysTimeArray: Array[0..1] of TSystemTime;
begin
  Perform(MCM_GETMONTHRANGE,Index,Integer(@SysTimeArray[0]));
  SystemTime2DateTime(SysTimeArray[0],Result.MinDate);
  SystemTime2DateTime(SysTimeArray[1],Result.MaxDate);
end;

procedure TMonthCalendar.SetSelChange(const Value: TOnEvent);
begin
  fOnSelChange := Value;
  AttachProc(@WndProcCalSelChange)
end;

function TMonthCalendar.GetOnGetMInfo: TOnGetMonthInfo;
begin
  if Assigned(CustomData) then Result := PMonthCalData(CustomData).fOnGetMonthInfo
  else Result := nil
end;

procedure TMonthCalendar.SetOnGetMInfo(const Value: TOnGetMonthInfo);
begin
  if Assigned(Value) then begin
    PMonthCalData(CustomData).fOnGetMonthInfo := Value;
    AttachProc(WndProcCalDayState);
    UpdateWndStyles;
    Style := Style or MCS_DAYSTATE;
  end else PMonthCalData(CustomData).fOnGetMonthInfo := nil;
end;

function TMonthCalendar.GetOnSelect: TOnSelect;
begin
  if Assigned(CustomData) then Result := PMonthCalData(CustomData).fOnSelect
  else Result := nil
end;

procedure TMonthCalendar.SetOnSelect(const Value: TOnSelect);
begin
  if Assigned(Value) then begin
    PMonthCalData(CustomData).fOnSelect := Value;
    AttachProc(WndProcCalSelChange);
  end else PMonthCalData(CustomData).fOnGetMonthInfo := nil;
end;

function TMonthCalendar.HitTest(const X, Y: Integer;
  var HitDate: TDateTime): Integer;
var
  HTI: TMCHitTestInfo;
begin
  FillChar(HTI,SizeOf(HTI),0);
  HTI.cbSize := SizeOf(TMCHitTestInfo);
  HTI.pt.X := X;
  HTI.pt.Y := Y;
  Result := Perform(MCM_HITTEST,0,Integer(@HTI));
  SystemTime2DateTime(HTI.st,HitDate);
end;

function TMonthCalendar.GetMaxDate: TDateTime;
begin
  GetRange();
  Result := PMonthCalData(fCustomData).fMaxDate
end;

function TMonthCalendar.GetMinDate: TDateTime;
begin
  GetRange();
  Result := PMonthCalData(fCustomData).fMinDate
end;

procedure TMonthCalendar.SetMaxDate(const Value: TDateTime);
begin
  with PMonthCalData(fCustomData)^ do begin
    fMaxDate := Value;
    SetRange(fMinDate,fMaxDate);
  end;
end;

procedure TMonthCalendar.SetMinDate(const Value: TDateTime);
begin
  with PMonthCalData(fCustomData)^ do begin
    fMinDate := Value;
    SetRange(fMinDate,fMaxDate);
  end;
end;

function TMonthCalendar.GetSelEnd: TDateTime;
begin
  GetSelRange();
  Result := PMonthCalData(fCustomData).fSelEnd;
end;

function TMonthCalendar.GetSelStart: TDateTime;
begin
  GetSelRange();
  Result := PMonthCalData(fCustomData).fSelStart;
end;


procedure TMonthCalendar.SetSelEnd(const Value: TDateTime);
begin
  SetSelRange(PMonthCalData(fCustomData).fSelStart,Value);
end;

procedure TMonthCalendar.SetSelStart(const Value: TDateTime);
begin
  SetSelRange(Value,PMonthCalData(fCustomData).fSelEnd);
end;

procedure TMonthCalendar.SetBoldDays;
begin
  Perform(MCM_SETDAYSTATE,Length(MonthBoldInfo),Integer(@MonthBoldInfo[0]));
end;

end.




